==> spring-boot-ext-1.2.63-sources.jar: cn/com/duibaboot/ext/autoconfigure/cloud/netflix/ribbon/RibbonPing.java <==
package cn.com.duibaboot.ext.autoconfigure.cloud.netflix.ribbon;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.LoggerContext;
import com.netflix.appinfo.InstanceInfo;
import com.netflix.client.config.IClientConfig;
import com.netflix.loadbalancer.AbstractLoadBalancerPing;
import com.netflix.loadbalancer.BaseLoadBalancer;
import com.netflix.loadbalancer.Server;
import com.netflix.niws.loadbalancer.DiscoveryEnabledServer;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpResponse;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.DefaultConnectionKeepAliveStrategy;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.protocol.HttpContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.MalformedURLException;

/**
 * 使用自定义的ping组件，每隔3秒访问服务器/monitor/check接口，如果不通或者返回不是OK，则标记为dead状态，去除对该服务器的流量（官方的只判断了从eureka取下来的状态是否UP，存在很大延迟.）
 *
 * "Ping" Discovery Client
 * i.e. we dont do a real "ping". We just assume that the server is up if Discovery Client says so
 * @author stonse
 *
 */
public class RibbonPing extends AbstractLoadBalancerPing {

    private static final Logger logger = LoggerFactory.getLogger(RibbonPing.class);

    static{
        LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
        loggerContext.getLogger(RibbonPing.class).setLevel(Level.INFO);
    }

    private static final CloseableHttpClient httpClient = HttpClientBuilder.create()
            .setDefaultRequestConfig(RequestConfig.custom().setConnectTimeout(100).setSocketTimeout(100).setConnectionRequestTimeout(100).build())
            .setMaxConnPerRoute(1)
//            .setMaxConnTotal(1000)
            .setUserAgent("RibbonPing")
            .disableAutomaticRetries()//禁止重试
            .disableCookieManagement()
            .useSystemProperties()//for proxy
//            .setConnectionTimeToLive()
            .setKeepAliveStrategy(new DefaultConnectionKeepAliveStrategy(){
                @Override
                public long getKeepAliveDuration(HttpResponse response, HttpContext context) {
                    long time = super.getKeepAliveDuration(response, context);
                    if(time == -1){
                        time = 30000;//链接最多空闲30秒
                    }
                    return time;
                }
            })
            .build();

    BaseLoadBalancer lb = null;

    public RibbonPing() {
    }

    public BaseLoadBalancer getLb() {
        return lb;
    }

    /**
     * Non IPing interface method - only set this if you care about the "newServers Feature"
     * @param lb
     */
    public void setLb(BaseLoadBalancer lb) {
        this.lb = lb;
    }

    public boolean isAlive(Server server) {
        boolean isAlive = true;
        if (server!=null && server instanceof DiscoveryEnabledServer){
            DiscoveryEnabledServer dServer = (DiscoveryEnabledServer)server;
            InstanceInfo instanceInfo = dServer.getInstanceInfo();
            if (instanceInfo!=null){
                InstanceInfo.InstanceStatus status = instanceInfo.getStatus();
                if (status!=null){
                    isAlive = status.equals(InstanceInfo.InstanceStatus.UP);
                }
            }

            Exception exception = null;
            String monitorCheckResult = null;
            if(isAlive) {
//                String healthCheckUrl = dServer.getInstanceInfo().getHealthCheckUrl();
                String monitorCheckUrl = "http://"+dServer.getHostPort()+"/monitor/check";
                HttpGet httpGet = new HttpGet(monitorCheckUrl);
                try(CloseableHttpResponse resp = httpClient.execute(httpGet)) {
                    int statusCode = resp.getStatusLine().getStatusCode();
                    monitorCheckResult = IOUtils.toString(resp.getEntity().getContent());//正常会拿到OK，这行只是为了消费掉，好让httpClient归还连接
                    if (statusCode == 404) {
                        logger.error("server: {} has no http endpoint /monitor/check, please setup,get help from here: http://cf.dui88.com/pages/viewpage.action?pageId=5252873");
                        isAlive = true;
                    } else if (statusCode >= 500) {
                        isAlive = false;
                    }
                } catch (MalformedURLException e) {
                    logger.error(e.getMessage(), e);
                    isAlive = false;
                } catch (IOException e) {
//                    e.printStackTrace();
                    isAlive = false;
                    exception = e;
                }
            }

            if(server.isAlive() != isAlive){
                if(exception != null){
                    logger.info("[ping in ribbon] detect that {}[{}] is {}，exceptionClass:{}, errorMsg:{}", instanceInfo.getAppName(), server.getHostPort(), isAlive?"alive":"dead", exception.getClass().getName(), exception.getMessage());
                }else{
                    logger.info("[ping in ribbon] detect that {}[{}] is {}, [/monitor/check] result:{}", instanceInfo.getAppName(), server.getHostPort(), isAlive?"alive":"dead", monitorCheckResult);
                }
            }
        }
        //TODO 这里可能会有并发问题，官方Server.isAliveFlag字段上没有volatile属性
        //TODO 还有个问题，虽然设置为非alive，但是eureka的线程每隔30秒获取到serverList后，会把所有server设置为alive，这段时间可能会有问题?好像没有问题
        return isAlive;
    }

    @Override
    public void initWithNiwsConfig(IClientConfig clientConfig) {
    }

    public void destroy(){
        try {
            httpClient.close();
        } catch (IOException e) {
            logger.error("", e);
        }
    }

}

==> spring-boot-ext-1.2.71-sources.jar: cn/com/duibaboot/ext/autoconfigure/cloud/netflix/ribbon/RibbonPing.java <==
package cn.com.duibaboot.ext.autoconfigure.cloud.netflix.ribbon;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.LoggerContext;
import com.netflix.appinfo.InstanceInfo;
import com.netflix.client.config.IClientConfig;
import com.netflix.loadbalancer.AbstractLoadBalancerPing;
import com.netflix.loadbalancer.BaseLoadBalancer;
import com.netflix.loadbalancer.Server;
import com.netflix.niws.loadbalancer.DiscoveryEnabledServer;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpResponse;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.DefaultConnectionKeepAliveStrategy;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.protocol.HttpContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.MalformedURLException;

/**
 * 使用自定义的ping组件，每隔3秒访问服务器/monitor/check接口，如果不通或者返回不是OK，则标记为dead状态，去除对该服务器的流量（官方的只判断了从eureka取下来的状态是否UP，存在很大延迟.）
 *
 * "Ping" Discovery Client
 * i.e. we dont do a real "ping". We just assume that the server is up if Discovery Client says so
 * @author stonse
 *
 */
public class RibbonPing extends AbstractLoadBalancerPing {

    private static final Logger logger = LoggerFactory.getLogger(RibbonPing.class);

    static{
        LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
        loggerContext.getLogger(RibbonPing.class).setLevel(Level.INFO);
    }

    private static final CloseableHttpClient httpClient = HttpClientBuilder.create()
            .setDefaultRequestConfig(RequestConfig.custom().setConnectTimeout(100).setSocketTimeout(100).setConnectionRequestTimeout(100).build())
            .setMaxConnPerRoute(1)
            .setMaxConnTotal(1000)
            .setUserAgent("RibbonPing")
            .disableAutomaticRetries()//禁止重试
            .disableCookieManagement()
            .useSystemProperties()//for proxy
            .setKeepAliveStrategy(new DefaultConnectionKeepAliveStrategy(){
                @Override
                public long getKeepAliveDuration(HttpResponse response, HttpContext context) {
                    long time = super.getKeepAliveDuration(response, context);
                    if(time == -1){
                        time = 30000;//链接最多空闲30秒
                    }
                    return time;
                }
            })
            .build();

    BaseLoadBalancer lb = null;

    public RibbonPing() {
    }

    public BaseLoadBalancer getLb() {
        return lb;
    }

    /**
     * Non IPing interface method - only set this if you care about the "newServers Feature"
     * @param lb
     */
    public void setLb(BaseLoadBalancer lb) {
        this.lb = lb;
    }

    public boolean isAlive(Server server) {
        boolean isAlive = true;
        if (server!=null && server instanceof DiscoveryEnabledServer){
            DiscoveryEnabledServer dServer = (DiscoveryEnabledServer)server;
            InstanceInfo instanceInfo = dServer.getInstanceInfo();
            if (instanceInfo!=null){
                InstanceInfo.InstanceStatus status = instanceInfo.getStatus();
                if (status!=null){
                    isAlive = status.equals(InstanceInfo.InstanceStatus.UP);
                }
            }

            Exception exception = null;
            String monitorCheckResult = null;
            if(isAlive) {
//                String healthCheckUrl = dServer.getInstanceInfo().getHealthCheckUrl();
                String monitorCheckUrl = "http://"+dServer.getHostPort()+"/monitor/check";
                HttpGet httpGet = new HttpGet(monitorCheckUrl);
                try {
                    CloseableHttpResponse resp = httpClient.execute(httpGet);//不能调用resp.close()否则tcp长连接会关闭
                    int statusCode = resp.getStatusLine().getStatusCode();
                    monitorCheckResult = IOUtils.toString(resp.getEntity().getContent());//正常会拿到OK，
                    resp.getEntity().getContent().close();//这行让httpClient归还连接到连接池
                    if (statusCode == 404) {
                        logger.error("server: {} has no http endpoint /monitor/check, please setup,get help from here: http://cf.dui88.com/pages/viewpage.action?pageId=5252873");
                        isAlive = true;
                    } else if (statusCode >= 500) {
                        isAlive = false;
                    }
                } catch (MalformedURLException e) {
                    logger.error(e.getMessage(), e);
                    isAlive = false;
                } catch (IOException e) {
//                    e.printStackTrace();
                    isAlive = false;
                    exception = e;
                }
            }

            if(server.isAlive() != isAlive){
                if(exception != null){
                    logger.info("[ping in ribbon] detect that {}[{}] is {}，exceptionClass:{}, errorMsg:{}", instanceInfo.getAppName(), server.getHostPort(), isAlive?"alive":"dead", exception.getClass().getName(), exception.getMessage());
                }else{
                    logger.info("[ping in ribbon] detect that {}[{}] is {}, [/monitor/check] result:{}", instanceInfo.getAppName(), server.getHostPort(), isAlive?"alive":"dead", monitorCheckResult);
                }
            }
        }
        //TODO 这里可能会有并发问题，官方Server.isAliveFlag字段上没有volatile属性
        //TODO 还有个问题，虽然设置为非alive，但是eureka的线程每隔30秒获取到serverList后，会把所有server设置为alive，这段时间可能会有问题?好像没有问题
        return isAlive;
    }

    @Override
    public void initWithNiwsConfig(IClientConfig clientConfig) {
    }

    public void destroy(){
        try {
            httpClient.close();
        } catch (IOException e) {
            logger.error("", e);
        }
    }

}
